class AllMetadata:
    """A class to represent a search for all metadata."""

    pass


ALL_METADATA = AllMetadata()
"""A special value to pass to :meth:`VectorIndexClientAsync.search` to request all metadata fields."""
